/* $Id: sfsauthd.C,v 1.18 2002/08/21 13:54:47 max Exp $ */

/*
 *
 * Copyright (C) 2001 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "authdb.h"
#include "sfsauthd.h"
#include "parseopt.h"
#include "rxx.h"

ptr<sfspriv> myprivkey;
sfs_servinfo myservinfo;
str sfsauthcachedir;
sfs_hash myhostid;
vec<dbfile> dbfiles;
str sfsauthrealm = "";
vec<sfsauth_certpath> sfsauthcertpaths;

void
xfer::start ()
{
  static str xferpath;
  if (pid != -1)
    return;
  if (!xferpath)
    xferpath = fix_exec_path ("xfer");
  const char *av[] = { xferpath, "--", src, dst, NULL };
  if ((pid = aspawn (xferpath, av)) != -1)
    chldcb (pid, wrap (mkref (this), &xfer::stop));
}

inline str
path2name (str path)
{
  char hash[sha1::hashsize];
  sha1_hash (hash, path, path.len ());
  return armor64A (hash, sizeof (hash));
}

ptr<xfer>
mkcache (str src)
{
  assert (src[0] == '/');
  str hash = path2name (src);
  str srclink =  sfsauthcachedir << "/source." << hash;

  errno = 0;
  char buf[256];
  int n = readlink (srclink, buf, sizeof (buf) - 1);
  if (n >= 0) {
    buf[n] = '\0';
    if (src != buf) {
      unlink (srclink);
      errno = ENOENT;
    }
  }
  if (errno && (errno != ENOENT || symlink (src, srclink) < 0)) {
    warn << srclink << ": " << strerror (errno) << "\n";
    return NULL;
  }
  return New refcounted<xfer> (src, sfsauthcachedir << "/cache." << hash);
}

void
cache_refresh ()
{
  for (dbfile *dbp = dbfiles.base (); dbp < dbfiles.lim (); dbp++)
    if (dbp->refresh)
      (*dbp->refresh) ();
  delaycb (60, wrap (cache_refresh));
}

void
dbfile::mkpub ()
{
  if (pubfiles.empty ())
    return;
  vec<ref<authcursor> > acv;
  for (str *np = pubfiles.base (); np < pubfiles.lim (); np++) {
    ptr<lockfile> lf = lockfile::alloc (*np << ".lock", false);
    if (!lf)
      continue;
    if (ptr<authcursor> ac = authcursor_file_append::alloc (*np, 0444, lf))
      acv.push_back (ac);
  }
  if (acv.empty ())
    return;
  ptr<authcursor> ac = db->open ();
  if (!ac)
    return;
  for (ac->reset (); ac->next ();) {
    switch (ac->ae.type) {
    case SFSAUTH_USER:
      ac->ae.userinfo->pwauth = "";
      ac->ae.userinfo->privkey.setsize (0);
      break;
    default:
      break;
    }
    for (ref<authcursor> *acp = acv.base (); acp < acv.lim (); acp++) {
      (*acp)->ae = ac->ae;
      (*acp)->update ();
    }
  }
  for (ref<authcursor> *acp = acv.base (); acp < acv.lim (); acp++)
    (*acp)->commit ();
}

str
dbfile::strip_prefix (str name)
{
  assert (name);
  if (!prefix)
    return name;
  if (name.len () <= prefix.len () || name[prefix.len ()] != '/'
      || memcmp (name.cstr (), prefix.cstr (), prefix.len ()))
    return NULL;
  return name;
}

static void
parseconfig (str cf)
{
  parseargs pa (cf);
  bool errors = false;
  bool empty = true;
  int line;
  vec<str> av;

  myservinfo.set_sivers (sfs_release);
  myservinfo.cr7->release = sfs_release;
  myservinfo.cr7->host.hostname = ""; // XXX - in case inited before 
                                      // rpc_emptystr

  while (pa.getline (&av, &line)) {
    if (!strcasecmp (av[0], "hostname")) {
      if (av.size () != 2) {
	errors = true;
	warn << cf << ":" << line << ": usage: hostname name\n";
      }
      else if (myservinfo.cr7->host.hostname != "") {
	errors = true;
	warn << cf << ":" << line << ": hostname already specified\n";
      }
      else
	myservinfo.cr7->host.hostname = av[1];
    }
    else if (!strcasecmp (av[0], "keyfile")) {
      if (myprivkey) {
	errors = true;
	warn << cf << ":" << line << ": keyfile already specified\n";
      }
      else if (av.size () == 2) {
	str keyfile (av[1]);
	if (keyfile[0] != '/')
	  keyfile = strbuf ("%s/", etc1dir) << keyfile;
	str key = file2wstr (keyfile);


	if (!key) {
	  errors = true;
	  warn << keyfile << ": " << strerror (errno) << "\n";
	  warn << cf << ":" << line << ": could not read keyfile\n";
	}
	else if (!(myprivkey = sfscrpt.alloc_priv (key, SFS_SIGN))) {
	  errors = true;
	  warn << cf << ":" << line << ": could not decode keyfile\n";
	}
      }
      else {
	errors = true;
	warn << cf << ":" << line << ": usage: keyfile path\n";
      }
    }
    else if (!strcasecmp (av[0], "userfile")) {
      static rxx pubrx ("^-pub=(.+)$", "i");
      static rxx prefixrx ("^-prefix=([\\w/]+)$", "i");
      static rxx uidid ("-uid=(\\d+)", "i");
      //static rxx uidnam ("-uid=([\w_\-]+)", "i");
      static rxx uidmap ("-uidmap=(\\d+)-(\\d+)+(\\d+)", "i");
      static rxx gidid ("-gid=(\\d+)", "i");
      //static rxx gidnam ("-gid=([\w_\-]+)", "i");
      static rxx gidmap ("-gidmap=(\\d+)-(\\d+)+(\\d+)", "i");
      dbfile *dbp = &dbfiles.push_back (dbfile ());
      av.pop_front ();
      for (; !av.empty () && av[0][0] == '-'; av.pop_front ())
	if (!strcasecmp (av[0], "-update"))
	  dbp->allow_update = true;
	else if (!strcasecmp (av[0], "-passwd"))
	  dbp->allow_unix_pwd = true;
	else if (!strcasecmp (av[0], "-admin"))
	  dbp->allow_admin = true;
	else if (pubrx.match (av[0])) {
	  str path = pubrx[1];
	  if (path[0] != '/')
	    path = strbuf () << etc1dir << "/" << path;
	  dbp->pubfiles.push_back (path);
	}
	else if (prefixrx.match (av[0])) {
	  if (dbp->prefix) {
	    errors = true;
	    warn << cf << ":" << line << ": extra -prefix\n";
	  }
	  dbp->prefix = prefixrx[1];
	}
	else if (uidid.match (av[0])) {
	  u_int32_t id = badid;
	  convertint (uidid[1], &id);
	  if (dbp->uidmap) {
	    errors = true;
	    warn << cf << ":" << line << ": only one uid/uidmap allowed\n";
	  }
	  dbp->uidmap = New refcounted<idmap_const> (id);
	}
	else if (uidmap.match (av[0])) {
	  u_int32_t min, max, offset;
	  convertint (uidid[1], &min);
	  convertint (uidid[2], &max);
	  convertint (uidid[3], &offset);
	  if (dbp->uidmap) {
	    errors = true;
	    warn << cf << ":" << line << ": only one uid/uidmap allowed\n";
	  }
	  dbp->uidmap = New refcounted<idmap_range> (min, max, offset);
	}
	else if (gidid.match (av[0])) {
	  u_int32_t id = badid;
	  convertint (gidid[1], &id);
	  if (dbp->gidmap) {
	    errors = true;
	    warn << cf << ":" << line << ": only one gid/gidmap allowed\n";
	  }
	  dbp->gidmap = New refcounted<idmap_const> (id);
	}
	else if (gidmap.match (av[0])) {
	  u_int32_t min, max, offset;
	  convertint (gidid[1], &min);
	  convertint (gidid[2], &max);
	  convertint (gidid[3], &offset);
	  if (dbp->gidmap) {
	    errors = true;
	    warn << cf << ":" << line << ": only one gid/gidmap allowed\n";
	  }
	  dbp->gidmap = New refcounted<idmap_range> (min, max, offset);
	}
	else {
	  errors = true;
	  warn << cf << ":" << line << ": unknown option " << av[0] << "\n";
	}
      if (av.size () == 0 || !av[0].len ()) {
	errors = true;
	warn << cf << ":" << line << ": missing file name\n";
      }
      else if (av.size () > 1) {
	errors = true;
	warn << cf << ":" << line << ": extra arguments after file name\n";
      }
      else if (dbp->allow_unix_pwd && dbp->uidmap) {
	errors = true;
	warn << cf << ":" << line
	     << ": same DB cannot have -passwd and -uidmap\n";
      }
      else {
	str path = av[0];
	if (path[0] != '/')
	  path = strbuf () << etc1dir << "/" << path;
	if (!dbp->allow_update) {
	  if (ptr<xfer> xp = mkcache (path)) {
	    dbp->refresh = wrap (xp, &xfer::start);
	    path = xp->dst;
	  }
	  else {
	    errors = true;
	    warn << cf << ":" << line
		 << ": cannot create cache file\n";
	  }
	}
	dbp->db = New refcounted<authdb_file> (path);
	if (!dbp->uidmap) {
	  dbp->uidmap = New refcounted<idmap_id> ();
	  if (!dbp->gidmap)
	    dbp->gidmap = New refcounted<idmap_id> ();
	}
	else if (!dbp->gidmap) {
	  warn << cf << ":" << line << ": assuming -gid=-1\n";
	  dbp->gidmap = New refcounted<idmap_const> (badid);
	}
	dbp->mkpub ();
      }
    }
    else if (!strcasecmp (av[0], "userdir")) {
      static rxx prefixrx ("^-prefix=([\\w/]+)$", "i");
      static rxx uidrange ("-uidrange=(\\d+)-(\\d+)", "i");
      static rxx gidid ("-gid=(\\d+)", "i");
      static rxx d ("-depth=(\\d+)", "i");

      u_int32_t uidstart = badid;
      u_int32_t uidend = badid;
      u_int32_t gid = badid;
      int depth = -1;

      dbfile *dbp = &dbfiles.push_back (dbfile ());
      av.pop_front ();
      for (; !av.empty () && av[0][0] == '-'; av.pop_front ())
	if (prefixrx.match (av[0])) {
	  if (dbp->prefix) {
	    errors = true;
	    warn << cf << ":" << line << ": extra -prefix\n";
	  }
	  dbp->prefix = prefixrx[1];
	}
	else if (uidrange.match (av[0])) {

	  if (uidstart != badid || uidend != badid) {
	    errors = true;
	    warn << cf << ":" << line << ": only one uidrange allowed\n";
	  } 
	  else {
	    convertint (uidrange[1], &uidstart);
	    convertint (uidrange[2], &uidend);
	  }

	}
	else if (gidid.match (av[0])) {
	  
	  if (gid != badid) {
	    errors = true;
	    warn << cf << ":" << line << ": only one gid allowed\n";
	  } 
	  else 
	    convertint (gidid[1], &gid);

	}
	else if (d.match (av[0])) {

	  if (depth != -1) {
	    errors = true;
	    warn << cf << ":" << line << ": only one depth allowed\n";
	  } 
	  else
	    convertint (d[1], &depth);

	}
	else {
	  errors = true;
	  warn << cf << ":" << line << ": unknown option " << av[0] << "\n";
	}
      if (av.size () != 2) {
	errors = true;
	warn << cf << ":" << line 
	     << ": missing directory name or cache file name\n"
	     << cf << ":" << line 
	     << ": usage: userdir [-uidrange=n-m] [gid=n] "
	     << "[-depth=n] dirpath cachepath\n";
      }
      else {
	str dirpath = av[0];
	str cachepath = av[1];
	if (dirpath[0] != '/' || cachepath[0] != '/') {
	  errors = true;
	  warn << cf << ":" << line 
	       << ": path names must be absolute\n";
	} else {
	  ptr<lockfile> lf;
	  if (!(lf = lockfile::alloc (cachepath << ".lock", wait))) {
	    warn << cf << ":" << line 
		 << ": lockfile " << cachepath << ".lock\n";
	    errors = true;
	  } else {

	    dbp->db = New refcounted<authdb_dir> (lf, dirpath, cachepath, 
						  uidstart, uidend, gid,
						  depth);
	    dbp->uidmap = New refcounted<idmap_id> ();
	    dbp->gidmap = New refcounted<idmap_id> ();
	    dbp->mkpub ();
	    dbp->allow_userdir_shell = true;
	  }
	}
      }
    }
    else {
      errors = true;
      warn << cf << ":" << line << ": unknown keyword " << av[0] << "\n";
    }
    empty = false;
  }

  // Default: Userfile -passwd -update -pub=<file>.pub <file>
  if (empty) {
    str userfile = sfsconst_etcfile_required ("sfs_users");
    strbuf pubuserfile = userfile << ".pub";
    dbfile *dbp = &dbfiles.push_back (dbfile ());
    dbp->pubfiles.push_back (pubuserfile);
    dbp->allow_update = true;
    dbp->allow_unix_pwd = true;
    dbp->db = New refcounted<authdb_file> (userfile);
    dbp->uidmap = New refcounted<idmap_id> ();
    dbp->gidmap = New refcounted<idmap_id> ();
    dbp->mkpub ();
  }

  if (!myservinfo.cr7->host.hostname.len ()) {
    str hostname = myname ();
    if (!hostname)
      fatal << "could not figure out my hostname\n";
    myservinfo.cr7->host.hostname = hostname;
  }
  if (!myprivkey) {
    str keyfile = sfsconst_etcfile ("sfs_host_key");
    if (!keyfile) {
      errors = true;
      warn << "cannot locate default file sfs_host_key\n";
    }
    else {
      str key = file2wstr (keyfile);
      if (!key) {
	errors = true;
	warn << keyfile << ": " << strerror (errno) << "\n";
      }
      else if (!(myprivkey = sfscrpt.alloc_priv (key, SFS_DECRYPT))) { 
	errors = true;
	warn << "could not decode " << keyfile << "\n";
      }
    }
  }
  if (errors)
    fatal ("errors in config file\n");

    

  myservinfo.cr7->host.type = SFS_HOSTINFO;
  myprivkey->export_pubkey (&myservinfo.cr7->host.pubkey);
  if (!sfs_mkhostid (&myhostid, myservinfo.cr7->host))
    fatal << "could not marshal my own hostinfo\n";
  myservinfo.cr7->prog = SFSAUTH_V2;
}

void
unixaccept (ptr<axprt_unix> x, const authunix_parms *aup)
{
  if (x) {
    vNew authclnt (axprt_crypt::alloc (x->reclaim ()), aup);
  }
}

ptr<axprt_stream>
cloneaccept (bool primary, int fd)
{
  if (fd < 0) {
    if (primary)
      fatal ("EOF from sfssd\n");
    return NULL;
  }
  tcp_nodelay (fd);
  ref<axprt_crypt> x = axprt_crypt::alloc (fd);
  vNew authclnt (x);
  return x;
}

void
clonegen (int fd)
{
  if (fd >= 0)
    if (!cloneserv (fd, wrap (cloneaccept, false))) {
      warn ("clonegen: cloneserv: %m\n");
      close (fd);
    }
}

static void
usage ()
{
  warnx << "usage: " << progname << " [-u sockpath] [-f configfile]\n";
  exit (1);
}

int
main (int argc, char **argv)
{
  setprogname (argv[0]);

  // XXX debug code
#if 0
  str key = file2wstr("/etc/sfs/sfs_host_key");
  ptr<rabin_priv> xx = import_rabin_priv (key , NULL);
  sfs_rabin_priv *p = new sfs_rabin_priv (xx, SFS_SIGN);
  sfspriv *p2 = New sfs_rabin_priv (xx, SFS_SIGN);
  ptr<sfspriv> p3 = New refcounted<sfs_rabin_priv> (xx, SFS_SIGN);
  exit (0);
#endif

  str configfile;
  str upath;
  int ch;
  while ((ch = getopt (argc, argv, "u:f:")) != -1)
    switch (ch) {
    case 'u':
      upath = optarg;
      break;
    case 'f':
      configfile = optarg;
      break;
    case '?':
    default:
      usage ();
    }

  argc -= optind;
  argv += optind;
  if (argc > 0)
    usage ();

  sfsauth_dbrec rec (SFSAUTH_USER);
  rec.userinfo->name = "logname";
  rec.userinfo->id = 1234;
  rec.userinfo->vers = 1;
  rec.userinfo->gid = 1234;
  rec.userinfo->pubkey.set_type (SFS_RABIN);
  rec.userinfo->audit = "changed by no one";

  warn ("version %s, pid %d\n", VERSION, getpid ());

  sfsconst_init ();

  sfs_suidserv ("authd", wrap (unixaccept));

  if (!runinplace) {
    if (!sfsauthcachedir)
      sfsauthcachedir = sfsdir << "/authdb";
    mksfsdir (sfsauthcachedir, 0755, NULL, 0);
  }
  else {
    if (!sfsauthcachedir)
      sfsauthcachedir = sfsdir << "/authdb";
    struct stat sb;
    if (stat (sfsdir, &sb) >= 0) {
      if (pid_t pid = fork ())
	waitpid (pid, NULL, 0);
      else {
	umask (0);
	setuid (sb.st_uid);
	mkdir (sfsauthcachedir, sb.st_mode & 0777);
	_exit (0);
      }
    }
  }
  if (!configfile)
    configfile = sfsconst_etcfile_required ("sfsauthd_config");
  parseconfig (configfile);
  if (!runinplace)
    mksfsdir (sfsauthcachedir, 0755, NULL, 0);
  else {
    struct stat sb;
    sb.st_uid = 0;
    stat (sfsdir, &sb);
    mksfsdir (sfsauthcachedir, 0755, NULL, sb.st_uid);
  }

  random_init_file (sfsdir << "/random_seed");

  warn << "serving " << sfs_hostinfo2path (myservinfo.cr7->host) << "\n";
#if 0
  if (sfsauthrealm.len () > 0)
    warn << "serving realm " << sfsauthrealm << "\n";
  authdbinit ();
#endif

  if (!cloneserv (0, wrap (cloneaccept, true)) && !upath)
    warn ("No sfssd detected, only accepting unix-domain clients\n");
  if (upath)
    sfs_unixserv (upath, wrap (clonegen));

  amain ();
}
