/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef CFSRW_RECORD_H
#define CFSRW_RECORD_H

#include "gcc_compat.h"
#include "record.h"
#include "cfsw_prot.h"  // cfsw_sattr

class attr_delta;
class write_delta;
class symlink_delta;
class inode_delta;
class dirop_delta;
class link_delta;
class unlink_delta;
class rename_delta;
class patch_delta;
class tag_delta;

extern deltap cfsw_deltap_parse (str s);

class cfsw_delta 
{
public:
  static const unsigned T_ATTR      = 10;
  static const unsigned T_WRITE     = 11;
  static const unsigned T_INODE     = 12;
  static const unsigned T_SYMLINK   = 13;
  static const unsigned T_LINK      = 14;
  static const unsigned T_UNLINK    = 15;
  static const unsigned T_RENAME    = 16;
  static const unsigned T_PATCH     = 17;
  static const unsigned T_TAG       = 18;
};

class attr_delta : public delta
{
private:
  bigint _object;
  cfsw_sattr _attr;

protected:
  attr_delta () : delta (cfsw_delta::T_ATTR) {}

public:
  attr_delta (const attr_delta& r) { *this = r; }
  attr_delta (bigint, cfsw_sattr);
  static ptr<attr_delta> alloc (str);

  bigint object () const { return _object; }
  cfsw_sattr* attr () { return &_attr; }
  str serialize () const;
};

class inode_delta : public delta
{
private:
  cfsw_ftype _object_type;
  bigint _object;
  cfsw_sattr _attr;

protected:
  inode_delta () : delta (cfsw_delta::T_INODE) {}

public:
  inode_delta (const inode_delta& r) { *this = r; }
  inode_delta (cfsw_ftype);
  inode_delta (cfsw_ftype, cfsw_sattr);
  inode_delta (cfsw_ftype, bigint);
  inode_delta (cfsw_ftype, bigint, cfsw_sattr);
  static ptr<inode_delta> alloc (str);

  cfsw_ftype object_type () const { return _object_type; }
  bigint object () const { return _object; }
  cfsw_sattr* attr () { return &_attr; }
  str serialize () const;
};

class symlink_delta : public delta
{
private:
  bigint _file;
  str _dest;

protected:
  symlink_delta () : delta (cfsw_delta::T_SYMLINK) {}

public:
  symlink_delta (const symlink_delta& r) { *this = r; }
  symlink_delta (bigint, str);
  static ptr<symlink_delta> alloc (str);

  bigint file () const { return _file; }
  str dest () const { return _dest; }

  str serialize () const;
};

class patch_delta : public delta
{
private:
  bigint _file;
  str _patch;

protected:
  patch_delta () : delta (cfsw_delta::T_PATCH) {}

public:
  patch_delta (bigint, const str&);
  static ptr<patch_delta> alloc (str);
  str serialize () const;

  bigint file () const { return _file; }
  const str& patch () const { return _patch; }
};

class write_delta : public delta
{
private:
  bigint _file;
  vec<bigint> _direct;
  bigint _indirect;
  bigint _double_indirect;
  bigint _triple_indirect;
  uint64 _size;

protected:
  write_delta () : delta (cfsw_delta::T_WRITE) {}

public:
  write_delta (const write_delta& r) { *this = r; }
  write_delta (bigint, const vec<bigint> &,
                    bigint, bigint, bigint, uint64);
  static ptr<write_delta> alloc (str);

  bigint file () const { return _file; }
  
  unsigned ndirects () const { return _direct.size (); }
  bigint direct (int i) const { return _direct [i]; }

  bigint indirect () const { return _indirect; }
  bigint double_indirect () const { return _double_indirect; }
  bigint triple_indirect () const { return _triple_indirect; }

  uint64 size () const { return _size; }

  str serialize () const;
};

class dirop_delta : public delta
{
protected:
  bigint _object;
  cfsw_ftype _object_type;
  bigint _dir;
  str _name;
  
  dirop_delta (int rtype, bigint object, cfsw_ftype object_type, 
                    bigint dir, str name);
  dirop_delta (int rtype) : delta (rtype) {}

public:
  virtual str serialize () const = 0;
  
  bigint dir () const { return _dir; }
  bigint object () const { return _object; }
  str name () const { return _name; }
  cfsw_ftype object_type () const { return _object_type; }
};

class link_delta : public dirop_delta
{
protected:
  link_delta () : dirop_delta (cfsw_delta::T_LINK) {}

public:
  link_delta (const link_delta &r)
    : dirop_delta (cfsw_delta::T_LINK) { *this = r; }
  link_delta (bigint object, cfsw_ftype object_type,
                   bigint dir, str name);
  static ptr<link_delta> alloc (str);
  str serialize () const;
};

class unlink_delta : public dirop_delta
{
protected:
  unlink_delta () : dirop_delta (cfsw_delta::T_UNLINK) {}

public:
  unlink_delta (const unlink_delta &r)
    : dirop_delta (cfsw_delta::T_UNLINK) { *this = r; }
  unlink_delta (bigint object, cfsw_ftype object_type,
                     bigint dir, str name);
  static ptr<unlink_delta> alloc (str);
  str serialize () const;
};

class rename_delta : public dirop_delta
{
private:
  bigint _from_dir;
  str _from_name;
  bool _target_exists;
  bigint _target;

protected:
  rename_delta () : dirop_delta (cfsw_delta::T_RENAME) {}

public:
  rename_delta (const rename_delta &r)
    : dirop_delta (cfsw_delta::T_RENAME) { *this = r; }
  rename_delta (bigint object, cfsw_ftype object_type,
                     bigint from_dir, str from_name,
                     bigint dir, str name, bigint target);
  rename_delta (bigint object, cfsw_ftype object_type,
                     bigint from_dir, str from_name,
                     bigint dir, str name);
  static ptr<rename_delta> alloc (str);
  str serialize () const;
  
  bigint from_dir ()  const { return _from_dir; }
  str from_name ()     const { return _from_name; }
  bool target_exists ()    const { return _target_exists; }
  bigint target () const { return _target; }
};

inline dirop_delta *
dirop_record (ptr<delta> r)
{
  if (r->rtype() == cfsw_delta::T_LINK)
    return r->TEMPLATE record<link_delta> ();
  else if (r->rtype() == cfsw_delta::T_UNLINK)
    return r->TEMPLATE record<unlink_delta> ();
  else if (r->rtype() == cfsw_delta::T_RENAME)
    return r->TEMPLATE record<rename_delta> ();
  return 0;
}

class tag_delta : public delta
{
private:
  bigint _commit;
  str _tag;

protected:
  tag_delta () : delta (cfsw_delta::T_TAG) {}

public:
  tag_delta (bigint, const str&);
  static ptr<tag_delta> alloc (str);
  str serialize () const;

  bigint commit () const { return _commit; }
  const str& tag () const { return _tag; }
};

#endif

