#include "pk_migrate.h"
#include "replica_manager.h"
#include <merkle_misc.h>
#include <blacklist.h>
#include <locationtable.h>

pk_migrate::pk_migrate(dhashcli *cli, ptr<vnode> host_node, 
		       ptr<dhash_blockstore> storage, ptr<keyhash_idx> kidx, 
		       int interval, int minreps, int maxreps, int maxdown) :
  cli (cli),
  host_node (host_node),
  storage (storage),
  kidx (kidx),
  interval (interval), 
  minreps (minreps),
  maxreps (maxreps),
  maxdown (maxdown),
  num_errors (0),
  outstanding_rpcs (0)
{
  int jitter = uniform_random (0, interval);
  active_cb = delaycb (interval + jitter, wrap (this, &pk_migrate::migrate_next_proj));
}

void pk_migrate::stop() {
  if (active_cb) {
    timecb_remove(active_cb);
    active_cb = NULL;
  }
}

// Advance to next pk set.
void pk_migrate::migrate_next_proj() {
  if (outstanding_rpcs == 0) {
    num_errors = 0;
    ptr<keyhash_idx::offer_state> s = 
      kidx->offernext (host_node->my_pred()->id(), host_node->my_ID());

    if (s) {
      warn << "Migrating (" << s->keys.size() << ") pk blocks for project " 
	   << armor64A(s->projkey) << "\n";
      num_errors = 0;

      cli->lookup
	(s->projkey, s->projkey, wrap (this, &pk_migrate::migrate_to_replicas, s));

    } else
      warn << "Migrating nothing. s is null\n";
  
  } else
    warn << "still " << outstanding_rpcs << " outstanding pk_migrate rpcs\n";

  active_cb = delaycb(interval, wrap(this, &pk_migrate::migrate_next_proj));
}

void pk_migrate::migrate_to_replicas (ptr<keyhash_idx::offer_state> s,
				      dhash_stat err, vec<chord_node> succs, route r) {

  vec<ptr<location> > replicas;
  for(unsigned i=0; i<succs.size() && i<maxreps; i++)
    replicas.push_back(host_node->locations->lookup_or_create(succs[i]));

  for(unsigned i=0; i<replicas.size(); i++) 
    if (host_node->my_ID() == replicas[i]->id()) {
      warn << "  I am a replica, not migrating\n";
      return;
    }
  
  for(unsigned rep_index=0; rep_index<replicas.size(); rep_index++) {
    if (blacklist::only().deny_project(replicas[rep_index]->address(), s->projkey) )
      continue;
    
    // offer keys to willing replicas[rep_index]
    unsigned sent = 0;
    while(sent < s->keys.size()) {
      unsigned lim = s->keys.size() - sent;
      if (lim > PER_RPC_KEYS)
	lim = PER_RPC_KEYS;
      
      ref<dhash_offer_pks_res> res =
	New refcounted<dhash_offer_pks_res> (DHASH_OK);
      ref<s_dhash_offer_pks_arg> arg =
	New refcounted<s_dhash_offer_pks_arg> ();
      
      arg->keys.setsize(lim);
      arg->versions.setsize(lim);
      
      for (unsigned i=0; i<lim; i++) {
	arg->keys[i] = s->keys[sent + i]; 
	arg->versions[i] = s->vers[sent + i];
      }
      
      sent += lim;
      outstanding_rpcs ++;
      
      warn << "  migrating to replica " << replicas[rep_index]->id() << "\n";
      
      host_node->doRPC
	(replicas[rep_index], dhash_program_1, DHASHPROC_OFFER_PKS, arg, res,
	 wrap (this, &pk_migrate::migrate_offerpks_cb,
	       replicas[rep_index], s, res));
    }
  }
  
  // No one wanted to take this project's block. Either, they were unwilling
  // or we dont have enough replicas. If we have enough replicas, but they are
  // all unwilling, we can remove the block.
  if (outstanding_rpcs == 0 && replicas.size() >= minreps)
    remove_project(s);
}

void pk_migrate::migrate_offerpks_cb(ptr<location> dest, 
				     ptr<keyhash_idx::offer_state> s,
				     ref<dhash_offer_pks_res> res, 
				     clnt_stat err) {
  outstanding_rpcs--;
  if (err || !res || (res->status != DHASH_OK && res->status != DHASH_UNWILLING)) {
    warn << "  error during offer\n";
    num_errors++;
    
  } else {
    unsigned sent = 0;
    for (unsigned i=0; i<res->resok->keys.size (); i++) {
      bigint projkey;
      ptr<dbrec> d = kidx->lookup (res->resok->keys [i], projkey);
      if (d) {
        if (projkey != s->projkey) {
	  warn << "offered " << s->projkey << " but reply "
	       << res->resok->keys [i] << " is under "
	       << projkey << "\n";
	  num_errors++; // (remote host is acting strange. dont remove the blocks later?)
        }
        outstanding_rpcs ++;
	sent ++;
        cli->sendblock
	  (dest, blockID (res->resok->keys [i], DHASH_KEYHASH, DHASH_BLOCK),
	   projkey, d,
	   wrap (this, &pk_migrate::send_block_cb, res->resok->keys[i], s));
	
      } else
	warn << "offer reply " << res->resok->keys [i] << " does not exist\n";
    }

    warn << "offer returned. Host " << dest->id() << " accepted "
	 << sent << " blocks\n";

    if (outstanding_rpcs == 0 && num_errors == 0) {
      warn << " no blocks accepted, removing project " << armor64A(s->projkey) << "\n";
      remove_project(s);
    } 
  }
}

  
void pk_migrate::send_block_cb(chordID k, ptr<keyhash_idx::offer_state> s, 
			       dhash_stat err, bool present) {
  outstanding_rpcs--;
  if (err) {
    num_errors++;
    warn << "  error sending block " << armor64A(k) << "\n";
  }
  
  if (outstanding_rpcs == 0 && num_errors == 0)
    // the blocks were all successfully offered/migrated, delete it now.
    remove_project(s);
}



void pk_migrate::remove_project(ptr<keyhash_idx::offer_state> s) {
  warn << "  removing project: " << armor64A(s->projkey) << "\n";

  for(unsigned i=0; i<s->keys.size(); i++) {
    warn << "    " << armor64A(s->keys[i]) << "\n";
    kidx->remove(s->keys[i]);
    storage->remove_pkblock(id2dbrec(s->keys[i]));
  }
}
