
#ifndef _CHUNKING_H_
#define _CHUNKING_H_

// we keep P(t), x, and K the same for whole file system, so two equivalent
// files would have the same breakmarks. for string A, fingerprint of A is
//
//   f(A) = A(t) mod P(t)
//
// we create breakmarks when
//
//   f(A) mod K = x
//
// if we use K = 8192, the average chunk size is 8k. we allow multiple K
// values so we can do multi-level chunking.

#include "vec.h"
#include "sha1.h"
#include "sfs_prot.h"
#include "rabinpoly.h"

#define FINGERPRINT_PT  0xbfe6b8a5bf378d83LL
#define BREAKMARK_VALUE 0x04

class chunk;

u_int64_t fingerprint(const unsigned char *data, size_t count);
int chunk_data (vec<chunk *>& cvp, const unsigned char *data, size_t count);
int chunk_file (vec<chunk *>& cvp, const char *path);

class Chunker {
private:
  struct prefetched_buffer {
    unsigned char *data;
    uint64 off;
    size_t size;
    prefetched_buffer (const unsigned char *d, uint64 o, size_t s) {
      data = New unsigned char[s];
      memmove (&data[0], d, s);
      off = o;
      size = s;
      next = 0;
    }
    struct prefetched_buffer *next;
  };

  window _w;
  size_t _last_pos;
  size_t _cur_pos;
  struct prefetched_buffer *_pfb;
  
  unsigned char *_hbuf; 
  unsigned int _hbuf_size;
  unsigned int _hbuf_cursor;

  vec<chunk *> _cv;
  void handle_hash(const unsigned char *data, size_t size);

public:
  Chunker();
  ~Chunker();

  size_t cur_pos () const { return _cur_pos; }
  void stop();
  void chunk_data (const unsigned char *data, size_t size);
  void chunk_data (const unsigned char *data, uint64 off, size_t size);

  const vec<chunk*>& chunk_vector() { return _cv; }
  void copy_chunk_vector (vec<chunk*>&);

  static unsigned keep_buf;
  static unsigned min_chunk_size;
  static unsigned max_chunk_size;
  static unsigned avg_chunk_size;
  static unsigned min_size_suppress;
  static unsigned max_size_suppress;
};

class chunk
{
private:
  off_t _pos;
  size_t _count;
  sfs_hash _hash;
  str _buf;

  void
  compute_hash (unsigned char *data, unsigned count)
  {
    sha1_hash (_hash.base(), data, count); 
  }

public:
  chunk (off_t p, size_t s, sfs_hash h)
    : _pos (p), _count (s), _hash(h), _buf (0) {}
  
  chunk (off_t p, size_t s, unsigned char *data)
    : _pos (p), _count (s)
    {
      if (Chunker::keep_buf)
	_buf = str ((const char*) data, (unsigned int)s);
      else
	_buf = 0;
      compute_hash (data, s);
    }

  size_t count () const { return _count; }
  off_t pos () const { return _pos; }
  str buf () const { return _buf; }

  sfs_hash hash () const { return _hash; }
  bigint hashn () const { 
    bigint n;
    mpz_set_rawmag_be (&n, _hash.base (), sha1::hashsize);
    return n;
  }
};

inline void
Chunker::copy_chunk_vector (vec<chunk*>& cvp)
{
  cvp.setsize(_cv.size());
  for (unsigned i=0; i<_cv.size(); i++)
    cvp[i] = New chunk(*(_cv[i]));
}


#endif /*_CHUNKING_H_*/

