/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/time.h>
#include "stdinc.h"
#include "parseopt.h"
#include "rxx.h"

#include "ns.h"
#include "record.h"
#include "memberlist.h"
#include "projinterface.h"
#include "misc_utils.h"

int ns_debug = 0;
int cfs_debug = 0;
bool userblock = false;
bool muststore = false;

const int MEMBERLIST = 1;
const int USERBLOCK  = 2;
int mode = MEMBERLIST;

ptr<ns> nsobj = 0;
ptr_block* _ptr_block;
ptr<sfspriv> _sk;
static int _work = 0;
  
static str addkey = 0;
str rule = 0;
static str name = "";
static bigint newkey;

void
local_exit (int err) {
  nsobj = NULL;
  exit (err);
}

void
create_block () 
{
  switch (mode) {
  case MEMBERLIST:
    _ptr_block = New memberlist;
    break;
  case USERBLOCK:
    _ptr_block = New ptr_block;
    break;
  }
}

void
store_cb (bool print, ns_stat status, bigint id)
{
  _work--;
  if (status != NS_OK)
    warn << "failed to store block\n";
  else if (print) {
    char raw[20];
    mpz_get_rawmag_be (raw, sizeof (raw), &id);
    str xx = armor64A (raw, sizeof (raw));
    strbuf out;
    out << xx << "\n";
    out.tosuio ()->output (1);
  }
  if (_work == 0)
    local_exit (0);
}

void
store_block ()
{
  _work++;
  _ptr_block->store (_sk, wrap (store_cb, true));
}

void
modify_memberlist (log_status status)
{
  if (status) {
    warn << "cannot load project block\n";
    local_exit (1);
  }

  //warn << "ptrblock (" << addkey << ") has " << _ptr_block->nentries() << " entries\n";
  //_ptr_block->print();

  if (addkey) {
    if (!_ptr_block->add_entry (newkey, name)) {
      warn << "cannot add entry, duplicate name?\n";
      local_exit (1);
    }
  }
  if (rule && mode == MEMBERLIST) {
    memberlist* ml = static_cast<memberlist*> (_ptr_block);
    ml->set_notify (rule);
  }
  _work++;
  _ptr_block->store (_sk, wrap (store_cb, true));
}

void
got_memberlist (bool verbose, log_status status)
{
  if (status) {
    warn << "cannot load project block\n";
    local_exit (1);
  }

  strbuf out;
  
  if (mode == MEMBERLIST) {
    memberlist* ml = static_cast<memberlist*> (_ptr_block);
    str n = ml->get_notify ();
    if (verbose && n && n != "") 
      out << "notify: " << n << "\n";
  }
  
  for (unsigned i=0; i<_ptr_block->nentries (); i++) {
    if (verbose)
      out << "member: " << _ptr_block->get_label (i) << " ("
	  << armor64A (_ptr_block->get_ptr (i)) << ")\n";
    else {
      bigint n = _ptr_block->get_ptr (i);
      char raw[20];
      mpz_get_rawmag_be (raw, sizeof (raw), &n);
      str s = armor64A (raw, sizeof (raw));
      out << s << "\n";
    }
  }
  out.tosuio ()->output (1);
  delete (_ptr_block);
  local_exit (0);
}

void
usage ()
{
  warnx << "usage: " << progname << " projectid\n"
        << "         prints the public keys in the project\n";
  warnx << "       " << progname << " -v projectid\n"
        << "         prints the public keys and Commit record hashes\n";

  warnx << "       " << progname << " -s adminkey\n"
        << "         creates a new project using <adminkey>\n";
  warnx << "       " << progname << " -s adminkey -n name -a key projectid\n"
        << "         adds <key> to project using <adminkey>\n";
  warnx << "       " << progname << " -s adminkey -r rule projectid\n"
        << "         adds <rule> as the new notification rule\n";

  warnx << "       " << progname << " -u memberkey projectid\n"
	<< "         creates a new member block using <memberkey>\n";

  // We should be able to figure out the memberid from the memberkey right?
  //warnx << "       " << progname << " -u memberkey -n label -a logpk memberid\n" 
  //<< "         adds <logpk> to the member block using <memberkey>\n";

  local_exit (1);
}


void
read_heads_cb(bool verbose, ptr<projinterface> projinf, 
	      ptr<head_state> heads, log_status status) {
  strbuf out;

  for (unsigned i=0; i<projinf->nlogs(); i++) {
    if (verbose) {

      int owner_idx = projinf->get_log_owner (i);
      str owner = (owner_idx >= 0) ? 
	projinf->get_user_label (owner_idx) : str("unknown");
      
      out << "log: " << owner << " "<< projinf->get_log_label(i) 
	  << " (" << armor64A(projinf->get_log_ptr(i)) << ")\n";

    } else
      out << armor64A(projinf->get_log_ptr(i)) << "\n";
  }
  out.tosuio()->output(1);
  local_exit(0);
}

void
start_query (bigint projid, bool verbose, str server) {
  ptr<projinterface> x = New refcounted<projinterface>(projid);
  x->fetch_heads(wrap(read_heads_cb, verbose, x));
}

void
start (bigint ptr_blk_addr, bool verbose, bool newmemberlist, str server)
{
  create_block();
  
  if (newmemberlist)
    store_block ();
  
  else if (addkey || rule) {
    if (addkey) {
      str keystr;

      rxx rx0 ("pk:([0-9a-zA-Z+-]*)$");
      rxx rx1 ("([0-9a-zA-Z+-]*)$");

      if (rx0.search (addkey))
        keystr = rx0 [1];
      else if (rx1.search (addkey))
        keystr = rx1 [1];
      else
        keystr = addkey;

      str dearm = dearmor64A(keystr);
      if (dearm)
        mpz_set_rawmag_be(&newkey, dearm.cstr(), dearm.len());
      else {
        warnx << "pastprojctl: key hash invalid\n";
        local_exit(1);
      }
    }
    _ptr_block->read_block (ptr_blk_addr, wrap (&modify_memberlist));
  }

  else {
    _ptr_block->read_block (ptr_blk_addr, wrap (&got_memberlist, verbose));
  }
}


int
main (int argc, char **argv)
{
  bool verbose = false;
  bool newmemberlist = false;
  char *skfile = 0;
  bool query = true;

  setprogname (argv[0]);
  random_init ();

  argv++;
  argc--;

  if (argc < 1)
    usage();

  while ( !strcmp(argv[0], "-r") || 
	  !strcmp(argv[0], "-n") ||
	  !strcmp(argv[0], "-a") ||
	  !strcmp(argv[0], "-s") ||
	  !strcmp(argv[0], "-v") ||
	  !strcmp(argv[0], "-u") ) {

    if (!strcmp(argv[0], "-r")) {
      rule = str (argv[1], strlen (argv[1]));
      argv += 2;
      argc -= 2;
    } else if (!strcmp(argv[0], "-n")) {
      name = str (argv[1], strlen (argv[1]));
      argv += 2;
      argc -= 2;
    } else if (!strcmp(argv[0], "-s")) {
      skfile = argv[1];
      query = false;
      argv += 2;
      argc -= 2;
    } else if (!strcmp(argv[0], "-a")) {
      addkey = str (argv[1], strlen(argv[1]));
      argv += 2;
      argc -= 2;
    } else if (!strcmp(argv[0], "-v")) {
      verbose = true;
      argv += 1;
      argc -= 1;
    } else if (!strcmp(argv[0], "-u")) {
      skfile = argv[1];
      query = false;
      argv += 2;
      argc -= 2;
      mode = USERBLOCK;
    }
    if (argc == 0)
      break;
  }

  // This is a mess.  We have to clean this up big time.
  
  if (argc > 1 || (argc < 1 && skfile == NULL))
    usage();

  if (skfile && !addkey && (!name || name == "") && !rule) {
    newmemberlist = true; 
  } else {
    if ( argc < 1) {
      warn << "too few args\n";
      usage();
    }
    if ( argc > 1) {
      warn << "too many args\n";
      usage();
    }
  }

  if ((newmemberlist || addkey || rule) && !skfile) {
    warn << "missing key\n";
    usage ();
  }

  if (addkey && name == "") {
    warn << "missing member name\n";
    usage ();
  }

  if (rule) {
    rxx rx ("^([CcEeRr]) ([a-zA-Z]*) ([\\w\\W\\s]*)$");
    if (!rx.search (rule)) {
      warn << "rule must be of the format: [CcEeRr] action arg\n";
      usage ();
    }
  }

  // Figure out what block we're working with.
  bigint projid = 0;

  if (argc == 1) {
    char sym[1024];
    str dearm;
    rxx rx ("project:([0-9a-zA-Z+-]*)[/]*$");
    int r = readlink(argv[0], sym, 1024);
    if (r > 0) sym[r] = '\0';
    if (rx.search (argv[0]) || (r > 0 && rx.search (sym)))
      dearm = dearmor64A(rx [1]);
    else
      dearm = dearmor64A(argv [0]);
    if (dearm) 
      mpz_set_rawmag_be(&projid, dearm.cstr(), dearm.len());
    else {
      warnx << "pastprojctl: invalid project ID\n";
      local_exit(1);
    }
  }

  if (newmemberlist || addkey || rule) {
    str key = file2wstr (skfile);
    if (!key) {
      warn << skfile << ": " << strerror (errno) << "\n";
      local_exit (1);
    }
    else if (!(_sk = sfscrypt.alloc_priv (key, SFS_SIGN))) {
      warn << "could not decode " << skfile << "\n";
      warn << key << "\n";
      local_exit (1);
    }
    if (projid == 0)
      projid = ns::pkid (_sk);
  }

  assert (projid != 0);
  check_create_replicadir();
  nsobj = ns::alloc (projid);
  
  if (query) 
    nsobj->connect (wrap (start_query, projid, verbose));
  else if (mode == MEMBERLIST) 
    nsobj->connect (wrap (start, projid, verbose, newmemberlist));
  else
    nsobj->connect (wrap (start, ns::pkid(_sk), verbose, newmemberlist));
  amain ();
}

