/*
 *  Copyright (C) 2005  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "projinterface.h"

const static int debug = 0;
const static int ignore_missing_userblocks = 1;

projinterface::projinterface(bigint projectid) :
  _work(0), _status(LOG_OK), _projectid(projectid)
{
  _memberlist = New refcounted<memberlist>();
  _headstate = New refcounted<head_state>();
}

void
projinterface::fetch_heads(callback<void, ptr<head_state>, log_status>::ref cb) {
  _work++;
  _final_cb = cb;
  _headstate->clear();
  _memberlist->read_block (_projectid, wrap(this, &projinterface::read_memberlist_cb));
}

void
projinterface::read_memberlist_cb(log_status status) {
  _work--;
  if (status) {
    if (debug) warn << "Could not read memberlist\n";
    _status = (status == LOGERR_NOBLOCK) ? LOGERR_NOPROJBLOCK : status;
    check_done();
    return;
  }

  // Retrieve the user blocks.  This can be furthur optimized using checkpks
  if (debug) warn << "memberlist has " << _memberlist->nentries() << " elements\n";
  for(unsigned i=0; i<_memberlist->nentries(); i++) {
    

    ptr<ptr_block> p = New refcounted<ptr_block>();
    _userblocks.push_back(p);

    if (_memberlist->get_ptr(i) != 0) {
      if (debug) warn << "Retrieving memberblock for " << _memberlist->get_label(i) << "\n";
      _work++;
      p->read_block(_memberlist->get_ptr(i), 
		    wrap(this, &projinterface::read_userblock_cb, i));
    } else if (debug)
      warn << "No memberblock for " << _memberlist->get_label(i) << "\n";
      
  }
  check_done();
}

void
projinterface::read_userblock_cb(unsigned user_idx, log_status status) {
  _work--;
  if (status) {
    warn << "Could not read memberblock for: " 
	 << armor64A(_memberlist->get_ptr(user_idx)) << "\n";
    _status = (status == LOGERR_NOBLOCK) ? LOGERR_NOUSERBLOCK : status;
    check_done();
    return;
  }

  // Retrieve head pointers.  This can be optimized by waiting for all
  // the user blocks to finish loading and then using checkpks to find
  // updated head pointers.
  ptr<ptr_block> p = _userblocks[user_idx];
  for(unsigned j=0; j<p->nentries(); j++) {
    if (p->get_ptr(j) != 0) {
      if (debug) warn << "Retrieving log pointer at: " << armor64A(p->get_ptr(j)) << "\n";
      _work++;
      nsobj->retrieve_lptr(p->get_ptr(j), 
			   wrap(this, &projinterface::read_lptr_cb, p->get_ptr(j), user_idx, j));
    }
  }
  check_done();
}


void
projinterface::read_lptr_cb (bigint lptr_addr, unsigned i, unsigned j,
			     lptr u, long pkv, ns_stat status)
{
  _work--;
  if (!u) {
    warn << "cannot load head pointer " << armor64A (lptr_addr) << "\n";
    _status = LOGERR_NOCPTR;
    check_done();
    return;
  }

  /*
  if (_pjdebug) {
    u_int64_t t = getusec () - _fetcht;
    warn << "load_lptr: reply, +" << (long) t << "\n";
  }
  */

  if (debug) warn << "Got the lptr at " << armor64A(lptr_addr) 
		  << " (" << u->pw_seqn()
		  << ") -> " << armor64A(u->commit()) << "\n";

  // Remember what we just learned about the branch heads
  _headstate->pkv.add (lptr_addr, u->pw_seqn());
  add_lptr(lptr_addr, u, _userblocks[i]->get_label(j), pkv);
  
  if (_log_to_user.size() < j+1)
    for(unsigned q=0; q<(j+1-_log_to_user.size()); q++) 
      _log_to_user.push_back(-1);
  _log_to_user[j] = i;

  // We need to save the log ptr address, and a ptr to the commit record
  _headstate->commits.add (lptr_addr, u->commit());

  // Again, this can be optimized to not load the commit if 
  // we have it already.  Actually, we won't load it if it's 
  // in the cache right?  Do we even want to do this here?
  if (u->commit() != 0) {
    _work++;
    nsobj->retrieve_commit (u->commit(), wrap(this, &projinterface::load_commit_cb, 
					      lptr_addr, u, i, j));
  }
  check_done();
}

void
projinterface::load_commit_cb(bigint lptr_addr, lptr u, unsigned i, unsigned j,
			      crec c, ns_stat status) {
  _work--;

  if (debug) warn << "Got commit record " << armor64A(u->commit()) 
		  << " for log pointer " << armor64A(lptr_addr) << "\n";

  if (!c || status) {
    warn << "cannot load Commit record " << armor64A (u->commit())
	 << " for log " << armor64A (lptr_addr) << "\n";
    _status = LOGERR_NOCOMMIT;
    check_done();
    return;
  }
  
  check_done();
}

void
projinterface::check_done() {
  if (_work > 0) 
    return;

  if (!_status || 
      (ignore_missing_userblocks && _status == LOGERR_NOUSERBLOCK)) {
    //_headstate->print();
    _final_cb(_headstate, _status);
  }

  if (_status) {
    ptr<head_state> results = New refcounted<head_state>();
    _final_cb(results, _status);
    return;
  }
  
}

void
projinterface::add_lptr(bigint new_addr, lptr new_lptr, str new_label, 
			long new_dhashseq) {
  bool added = false;
  vec<bigint> addrs;
  vec<lptr>   lptrs;
  vec<str>    labels;
  vec<unsigned>   pw_seqns;
  vec<long>   dh_seqns;

  while(_headstate->lptrs.size() > 0) {
    if (new_addr < _lptr_addrs[0] && !added) {
      addrs.push_back(new_addr);
      lptrs.push_back(new_lptr);
      labels.push_back(new_label);
      dh_seqns.push_back(new_dhashseq); 
      added = true;
    } else {
      addrs.push_back(_lptr_addrs.pop_front());
      lptrs.push_back(_headstate->lptrs.pop_front());
      labels.push_back(_labels.pop_front());
      dh_seqns.push_back(_dhash_seqns.pop_front());
    }      
  }
  
  if (!added){
    addrs.push_back(new_addr);
    lptrs.push_back(new_lptr);
    labels.push_back(new_label);
    dh_seqns.push_back(new_dhashseq);
  }

  _headstate->lptrs = lptrs;
  _lptr_addrs = addrs;
  _labels = labels;
  _dhash_seqns = dh_seqns;
}

unsigned 
projinterface::nusers() const {
  return _memberlist->nentries();
}

bigint 
projinterface::get_user_id(unsigned i) const {
  return _memberlist->get_ptr(i);
}

str
projinterface::get_user_label(unsigned i) const {
  return _memberlist->get_label(i);
}

long
projinterface::get_userblk_version (unsigned i) const {
  return _userblocks[i]->get_blk_revision();
}

unsigned
projinterface::nlogs() const {
  return _lptr_addrs.size();
}

bigint
projinterface::get_log_ptr(unsigned i) const {
  return _lptr_addrs[i];
}

str
projinterface::get_log_label(unsigned i) const {
  return _labels[i];
}

long
projinterface::get_log_dhash_version (unsigned i) const {
  return _dhash_seqns[i];
}

int 
projinterface::get_log_owner (unsigned i) const {
  assert (i < _log_to_user.size());
  assert (_log_to_user[i] >= 0);
  return _log_to_user[i];
}
str
projinterface::get_notify() const {
  return _memberlist->get_notify();
}

long
projinterface::get_projblk_version () const {
  return _memberlist->get_blk_revision ();
}
