/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include "pastwatch.h"
#include "opts.h"
#include "io.h"

extern ptr<cfsrw> _cfsrw;

static void
noop_cb (cbi cb, int n)
{
  cb (n);
}

void
list_branches (int err = 0)
{
  strbuf out;
  const ptr<replica> r = _cfsrw->getreplica ();

  for (unsigned i=0; i < r->nbranches (); i++) {
    commit_info inf = r->branch_head_info(i);
    strbuf x;
    x << "branch \"" << r->branch_tag (i)
      << "\": head is " << armor64A (inf.rptr) << "\n";
    if (err)
      warn << x;
    else
      out << x;
  }
  if (!err)
    out.tosuio ()->output (1);
}

static void show_history (bigint c, cbi cb);

static void
retrieve_block_cb (bigint prev, unsigned level, cbi cb, ptr<ns_block> b)
{
  if (b) {
    str s (b->data, b->len);
    crec r = commit_record::alloc (s);
    if (r) {
      strbuf out;
      out << "date: " << myctime ((time_t) r->time ()) << "\n";
      if (r->msg () == "")
	out << "*** empty log message ***\n";
      else
	out << r->msg () << "\n";
      out.tosuio ()->output (1);
    }
  }

  if (level > 0)
    show_history (prev, cb);
  else
    delaycb (0, wrap (noop_cb, cb, 0));
}

static void
show_history (bigint c, cbi cb)
{
  ptr<btree_srv> b = _cfsrw->getreplica ()->btree ();
  btree_node x;
  if (!b->lookup_by_commit (c, x)) {
    warn << "cannot find " << armor64A (c) << " in branch tree\n";
    delaycb (0, wrap (noop_cb, cb, 0));
  }
  else {
    strbuf out;
    out << "----------------------------\n";
    out << "Version: ";
   
    bigint n;
    if (b->lookup_by_tag (x.proposed_name, n))
      out << x.proposed_name << " ";
    out << "(" << armor64A (c) << ")\n";
    out.tosuio ()->output (1);
    nsobj->retrieve_block
      (c, wrap (retrieve_block_cb, x.prev.rptr, x.level, cb));
  }
}

extern str get_tag_of_commit (bigint cn);
static void get_tag_info (vec<str> x, cbi cb);

static void
get_tag_commit (vec<str> x, cbi cb, str t, ptr<ns_block> b)
{
  if (b) {
    str s (b->data, b->len);
    crec r = commit_record::alloc (s);
    if (r) {
      strbuf out;
      out << "Symbolic tag: " << t << " -> version: " << r->proposed_name() 
	  << ", "<< myctime ((time_t) r->time ()) << "\n";
      out.tosuio ()->output (1);
    }
  }
  get_tag_info (x, cb);
}

static void
get_tag_info (vec<str> x, cbi cb)
{
  ptr<btree_srv> srv = _cfsrw->getreplica ()->btree ();
  if (x.size ()) {
    str t = x.pop_front ();
    bigint c;
    if (srv->lookup_by_tag (t, c)) {
      nsobj->retrieve_block (c, wrap (get_tag_commit, x, cb, t));
      return;
    }
  }
  cb (0);
}

void
admin (cbi cb)
{
  ptr<replica> r = _cfsrw->getreplica ();

  if (arg_command == cmd_latest) {
    list_branches ();
    delaycb (0, wrap (noop_cb, cb, 0));
  }

  else if (arg_command == cmd_history) {
    assert (arg_branch != "");
    strbuf out;
    out << "branch: head: "
	<< get_tag_of_commit (my_branch_head (r))
	<< ", tag: " << r->branch_tag (r->my_branch ()) << "\n";
    out.tosuio ()->output (1);
    show_history (my_branch_head (r), cb);
  }

  else if (arg_command == cmd_tags) {
    ptr<btree_srv> b = r->btree ();
    str s = b->get_tags ();
    if (s) {
      vec<str> xs;
      str t;
      xin x (s);
      while (x.rstr (t))
	xs.push_back (t);
      get_tag_info (xs, cb);
      return;
    }
    delaycb (0, wrap (noop_cb, cb, 0));
  }

  else {
    assert (0);
    delaycb (0, wrap (noop_cb, cb, 0));
  }
}

void find_commit_by_time
(unsigned t, bigint c, callback<void,bigint,int>::ref cb);

static void
fetch_commit_cb (bigint c, bigint prev, unsigned level, unsigned t,
                 callback<void,bigint,int>::ref cb, ptr<ns_block> b)
{
  if (b) {
    str s (b->data, b->len);
    crec r = commit_record::alloc (s);
    if (r->time () < t) { // found it
      cb (c, 0);
      return;
    }
  }
  if (level > 1)
    find_commit_by_time (t, prev, cb);
  else
    cb (c, 0);
}

void
find_commit_by_time (unsigned t, bigint br, callback<void,bigint,int>::ref cb)
{
  ptr<btree_srv> b = _cfsrw->getreplica ()->btree ();
  btree_node x;
  if (!b->lookup_by_commit (br, x)) {
    warn << "cannot find " << armor64A (br) << " in branch tree\n";
    cb (br, -1);
  }
  else
    nsobj->retrieve_block
      (br, wrap (fetch_commit_cb, br, x.prev.rptr, x.level, t, cb));
}

