/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "pastwatch.h"
#include "io.h"
#include "path.h"
#include "pdb.h"


extern void
read_file (int, bigint, bigint, callback<void, cfsrw_status>::ref cb);

extern ptr<cfsrw> _cfsrw;

class checkout_object : public virtual refcount
{
private:
  int work;
  int error;
  bigint cr_addr;
  bool update_pdb;
  unsigned outstanding;
  unsigned max_outstanding;

  struct checkout_dirent_t {
    bigint fh;
    str path;
    str cfspath;
  };
  typedef ptr<checkout_dirent_t> cdep_t;

  struct checkout_entry {
    cdep_t cde;
    ptr<cfsr_attr> attr;
    callback<void, int>::ptr cb;
    
    checkout_entry(){}
    checkout_entry(cdep_t _cde, ptr<cfsr_attr> _attr, cbi _cb) :
      cde(_cde), attr(_attr), cb(_cb) {  }
  };
  vec< ptr<checkout_entry> > queue;

  void enqueue (cdep_t cde, ptr<cfsr_attr> attr, cbi cb) {
    queue.push_back(New refcounted<checkout_entry>(cde, attr, cb));
    dequeue();
  }
  void dequeue () {
    while(outstanding < max_outstanding && queue.size()) {
      outstanding++;
      ptr<checkout_entry> e = queue.pop_front();
      read_one_file(e->cde, e->attr, e->cb);
    }
  }

  void
  read_reply (cdep_t cde, int fd, uint32 mode, cbi cb, cfsrw_status status)
  {
    work--;
    outstanding--; 
    dequeue();

    if (status) {
      error = -1;
      if (status) {
        close (fd);
        warn << "cannot download " << cde->path << "\n";
      }
      if (work == 0)
        cb (error);
      return;
    }
          
    chmod (cde->path, mode);
    
    close (fd);
    pastwatch_status ("U", cde->path);
    if (update_pdb)
      pmgr->schedule_add (cde->path, pfe (cde->fh, cr_addr));
  
    if (work == 0)
      cb (error);
  }

  void
  readlink_reply (cdep_t cde, cbi cb, cfsrcallp cc, cfsrw_status status)
  {
    work--;

    if (status) {
      error = -1;
      warn << "cannot download " << cde->path << "\n";
      if (work == 0)
        cb (error);
      return;
    }

    cfsr_readlink_res *res = static_cast<cfsr_readlink_res *> (cc->res ());
    str dest = res->dest;
   
    int r = symlink (dest.cstr (), cde->path);
    if (r < 0) {
      warn << "cannot create symbolic link " << cde->path << "\n";
      error = -1;
    }
    else {
      pastwatch_status ("S", cde->path);
      if (update_pdb)
	pmgr->schedule_add (cde->path, pfe (cde->fh, cr_addr));
    }
  
    if (work == 0)
      cb (error);
  }
  
  void
  readdir_reply (cdep_t cde, cbi cb, cfsrcallp cc, cfsrw_status status)
  {
    work--;

    if (status) {
      error = -1;
      warn << "cannot checkout directory " << cde->path << "\n";
      if (work == 0)
        cb (error);
      return;
    }
  
    cfsr_readdir_res *res = static_cast<cfsr_readdir_res *> (cc->res ());
    ptr<cfs_dirlist> dir = res->dir;
  
    rpc_ptr<cfs_rpc_dirent> de = dir->entries;
    uint64 cookie = 0;
  
    while (de) {
      cdep_t new_cde = New refcounted<checkout_dirent_t>;
      new_cde->path = make_path (cde->path, de->name);
      new_cde->cfspath = make_path (cde->cfspath, de->name);
      new_cde->fh = de->fh;
  
      work++;

      cfsrcallp c = New refcounted<cfsrcall>
	(CFSRPROC_GETATTR,
	 wrap (mkref(this), &checkout_object::getattr_reply, new_cde, cb));
      cfsr_getattr_arg *arg = static_cast<cfsr_getattr_arg *> (c->arg ());
      arg->object = de->fh;
      cfsr_call (cr_addr, c);

      cookie = de->cookie;
      de = de->nextentry;
    }
  
    if (!dir->eof) {
      work++;

      cfsrcallp c = New refcounted<cfsrcall> 
	(CFSRPROC_READDIR,
	 wrap (mkref(this), &checkout_object::readdir_reply, cde, cb));
      cfsr_readdir_arg *arg = static_cast<cfsr_readdir_arg *> (c->arg ());
      arg->dir = cde->fh;
      arg->cookie = cookie;
      arg->count = 2048;
      cfsr_call (cr_addr, c);
    }

    else if (work == 0)
      cb (error);
  }

  void
  read_one_file(cdep_t cde, ptr<cfsr_attr> attr, cbi cb) 
  {
    int fd = open
      (cde->path, O_CREAT | O_TRUNC | O_WRONLY, 0644);
    if (fd < 0) {
      error = -1;
      warn << "cannot open " << cde->path << " for writing\n";
      work--;
      outstanding--;
      dequeue();
      return;
    }
    read_file
      (fd, cde->fh, cr_addr,
       wrap (mkref(this), &checkout_object::read_reply,
	     cde, fd, attr->mode, cb));
  }

  void
  getattr_reply (cdep_t cde, cbi cb, cfsrcallp cc, cfsrw_status status)
  {
    work--;
    
    cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (cc->res ());
    ptr<cfsr_attr> attr = res->attr;

    if (status) {
      error = -1;
      warn << "cannot get attribute for " << cde->path << "\n";
      if (work == 0)
        cb (error);
      return;
    }

    if (attr->type == CFSREG) {
      work++;
      enqueue(cde, attr, cb);
    }
  
    else if (attr->type == CFSLNK) {
      work++;
      
      cfsrcallp c = New refcounted<cfsrcall> 
	(CFSRPROC_READLINK,
	 wrap (mkref(this), &checkout_object::readlink_reply, cde, cb));
      cfsr_readlink_arg *arg = static_cast<cfsr_readlink_arg *> (c->arg ());
      arg->file = cde->fh;
      cfsr_call (cr_addr, c);
    }
  
    else if (attr->type == CFSDIR) {
      if (mkdir (cde->path, 0755) < 0 && errno != EEXIST) {
        error = -1;
        warn << "cannot create directory " << cde->path << "\n";
        if (work == 0)
          cb (error);
        return;
      }
  
      if (update_pdb)
	pmgr->schedule_mkdir (cde->path, cde->cfspath);
  
      work++;
  
      cfsrcallp c = New refcounted<cfsrcall> 
	(CFSRPROC_READDIR,
	 wrap (mkref(this), &checkout_object::readdir_reply, cde, cb));
      cfsr_readdir_arg *arg = static_cast<cfsr_readdir_arg *> (c->arg ());
      arg->dir = cde->fh;
      arg->cookie = 0;
      arg->count = 2048;
      cfsr_call (cr_addr, c);
     
      pastwatch_status ("D", cde->path);
      if (update_pdb)
	pmgr->schedule_add (cde->path, pfe (cde->fh, cr_addr));
    }

    if (work == 0)
      cb (error);
  }
  
public:
  checkout_object (bigint fh, str path, str cfspath, bigint cr_addr,
                   cbi cb, bool update)
    : work (0), error (0), cr_addr (cr_addr), update_pdb (update),
      outstanding(0), max_outstanding(512)
  {
    cdep_t cde = New refcounted<checkout_dirent_t>;
    cde->fh = fh;
    cde->path = path;
    cde->cfspath = cfspath;
    work++;

    cfsrcallp c = New refcounted<cfsrcall> 
      (CFSRPROC_GETATTR,
       wrap (mkref(this), &checkout_object::getattr_reply, cde, cb));
    cfsr_getattr_arg *arg = static_cast<cfsr_getattr_arg *> (c->arg ());
    arg->object = cde->fh;
    cfsr_call (cr_addr, c);
  }
};


static bigint
get_my_cr_addr ()
{
  if (arg_tag[0] != "") {
    bigint n;
    dearmor64A_bigint (arg_tag[0], n);
    return n;
  }
  else
    return my_branch_head (_cfsrw->getreplica ());
}


void
export_one (bigint handle, str path, cbi cb)
  // does not update pdb
{
  bigint cr_addr = get_my_cr_addr ();
  ptr<checkout_object> a =
    New refcounted<checkout_object> (handle, path, "", cr_addr, cb, false);
}

void
checkout_one (bigint handle, str path, str cfspath, bigint cr_addr, cbi cb)
{
  ptr<checkout_object> a =
    New refcounted<checkout_object>
    (handle, path, cfspath, cr_addr, cb, !arg_noupdate);
}

static int
recursive_make_parent (str path, mode_t mode, vec<bigint>& fhs, bigint cr_addr)
{
  vec<str> names;
  splitpath (names, path);
  str p = "";
  int i = 0;
  while (names.size () > 1) {
    str dir = names.pop_front ();
    if (p == "")
      p = dir;
    else
      p << "/" << dir;
    int r;
    if ((r = mkdir (p, mode)) < 0 && errno != EEXIST)
      return -1;
    if (!r) { // successful
      pmgr->schedule_mkdir (p, p);
      bigint fh = fhs[i];
      pmgr->schedule_add (p, pfe (fh, cr_addr));
    }
    i++;
  }
  return 0;
}

static void
lookup_reply (str path, str cfspath, bigint cr_addr, cbi cb,
              ptr<vec<bigint> > fhs, cfsrw_status status)
{
  if (status) {
    warn << "cannot lookup " << cfspath << " in repository\n";
    cb (-1);
    return;
  }

  if (path == "") {
    path = cfspath;
    if (path[path.len()-1] == '/')
      path = substr(path, 0, path.len()-1);

    char* end = strrchr(path.cstr(), '/');
    if (end) {
      int pos = end - path.cstr();
      path = substr(path, pos + 1, path.len() - pos - 1);
    }
    
    recursive_make_parent (path, 0755, *fhs, cr_addr);
  }
  else {
    mkdir (path, 0755);
    struct stat sb;
    if (lstat (path, &sb) || !(sb.st_mode & S_IFDIR)) {
      warn << "cannot stat directory " << path << "\n";
      cb (-1);
      return;
    }
  }

  checkout_one (fhs->pop_back (), path, cfspath, cr_addr, cb);
}

void
checkout (bigint root, str path, str cfspath, cbi cb)
{
  bigint cr_addr = get_my_cr_addr ();
  cfsr_lookup_path
    (root, cfspath, wrap (lookup_reply, path, cfspath, cr_addr, cb));
}

