/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <errno.h>
#include <rxx.h>
#include "pastwatch.h"
#include "path.h"
#include "pdb.h"
#include "opts.h"
#include "penviron.h"

extern str get_btag_from_infofile (str path);

/*
 * we keep version data for files in the working directory in a db.
 * this db contains mapping from local file name to cfs file handle.
 * we also keep, for each entry, the address of the snapshot the file
 * was retrieved from, and the mtime after the file was copied into
 * the working directory.
 */

class pfe_factory : public adb_entry_factory {
public:
  ptr<adb_entry>
  parse_entry (const char *s)
  {
    ptr<pfe> e = New refcounted<pfe> ();
    e->parse_entry (s);
    return e;
  }
};

static pfe_factory *pfef = 0;

void
pfe::print_entry (strbuf &b)
{
  b << name << "/" << armor64A (fh) << "/"
    << armor64A (cr_addr) << "/"
    << mtime.sec << "." << mtime.nsec;
  if (sticky != 0)
    b << "/" << armor64A (sticky);
}

bool
pfe::parse_entry (const char *s)
{
  vec<str> n;
  split (&n, "/", s);

  if (n.size () == 4 || n.size () == 5) {
    name = n [0];
    if (!dearmor64A_bigint (n [1], fh))
      return false;
    if (!dearmor64A_bigint (n [2], cr_addr))
      return false;
    vec<str> m;
    split (&m, "\\.", n [3]);
    if (m.size () == 2) {
      mtime.sec = atoi (m [0].cstr ());
      mtime.nsec = atoi (m [1].cstr ());
    }
    else
      return false;
    if (n.size () == 5) {
      if (!dearmor64A_bigint (n [4], sticky))
        return false;
    }
    else
      sticky = 0;
    return true;
  }
  return false;
}

ptr<adb>
pdb_mgr::opendb (str dir)
{
  ptr<adb> *dbptr = adbhash [dir];
  if (dbptr) {
    assert (*dbptr);
    return *dbptr;
  }

  str r  = dir << "/" << PASTWATCH_REF;
  struct stat sb;
  if (lstat (r, &sb))
    return 0L;

  str dbpath = get_db_file (dir);
  ptr<adb> db = New refcounted <adb> ();
  if (db->opendb (dbpath, pfef, true)) {
    warn << "cannot open " << dbpath << "\n";
    return 0L;
  }

  adbhash.insert (dir, db);
  adbs.push_back (db);
  opened_dirs.push_back (dir);

  return db;
}

int
pdb_mgr::init ()
{
  if (!pfef)
    pfef = New pfe_factory ();

  add_path.clear ();
  add_pfe.clear ();
  mkdir_path.clear ();
  mkdir_cfsp.clear ();
  mkdir_renamed.clear ();
  remove_path.clear ();

  str h = pget_pwroot();
  if (mkdir (h, 0755) && errno != EEXIST) {
    perror ("cannot create pastwatch root directory");
    return -1;
  }
  return 0;
}

void
pdb_mgr::schedule_rename (str newdir, str cfsto)
{
  mkdir_path.push_back (newdir);
  mkdir_cfsp.push_back (cfsto);
  mkdir_renamed.push_back (true);
}

void
pdb_mgr::schedule_mkdir (str path, str cfspath)
{
  mkdir_path.push_back (path);
  mkdir_cfsp.push_back (cfspath);
  mkdir_renamed.push_back (false);
}

void
pdb_mgr::schedule_remove (str path)
{
  remove_path.push_back (path);
}

void
pdb_mgr::schedule_supersedes (str path, bigint n)
{
  supers_dirs.push_back (path);
  supers_commits.push_back (n);
}

void
pdb_mgr::schedule_add (str path, struct pfe e)
{
  add_path.push_back (path);
  add_pfe.push_back (e);
}

mtime_t
get_mtime (str path)
{
  mtime_t m;
  m.sec = m.nsec = 0;

  struct stat sb;
  if (lstat (path, &sb)) {
    perror ("cannot stat file");
    return m;
  }

#ifdef SFS_HAVE_STAT_ST_MTIMESPEC
  if (sb.st_mtimespec.tv_sec > sb.st_ctimespec.tv_sec ||
      (sb.st_mtimespec.tv_sec == sb.st_ctimespec.tv_sec &&
       sb.st_mtimespec.tv_nsec > sb.st_ctimespec.tv_nsec)) {
    m.sec = sb.st_mtimespec.tv_sec;
    m.nsec = sb.st_mtimespec.tv_nsec;
  }
  else {
    m.sec = sb.st_ctimespec.tv_sec;
    m.nsec = sb.st_ctimespec.tv_nsec;
  }
#else
  if (sb.st_mtime > sb.st_ctime ||
      (sb.st_mtime == sb.st_ctime && sb.st_mtime > sb.st_ctime)) {
    m.sec = sb.st_mtime;
    m.nsec = 0;
  }
  else {
    m.sec = sb.st_ctime;
    m.nsec = 0;
  }
#endif

  return m;
}

int
pdb_mgr::process (bigint cr_addr, str btag)
{
  if (!btag)
    btag = "";

  for (unsigned i = 0; i < mkdir_path.size (); i++) {
    assert (!arg_checkonly);

    str p = mkdir_cfsp [i];
    str d = mkdir_path [i] << "/" << PASTWATCH_REF;

    if (!mkdir_renamed [i]) {
      if (mkdir (d, 0755)) {
        warn << "cannot create " << d << "\n";
        return -1;
      }
    }

    str f = get_info_file (mkdir_path [i]);
    str m = p << "\n"
              << arg_logkeyfile << "\n"
	      << arg_projstr << "\n"
	      << arg_branch  << "\n"
	      << btag << "\n";

    if (!str2file (f, m, 0644, false)) {
      warn << "cannot write to " << f << "\n";
      return -1;
    }

    f = get_ops_file (mkdir_path [i]);
    int fd = open (f, O_CREAT | O_TRUNC | O_WRONLY, 0644);
    if (fd < 0) {
      warn << "cannot create " << f << "\n";
      return -1;
    }
    close (fd);
  
    str dbp = get_db_file (mkdir_path [i]);
    adb db;
    if (int err = db.opendb (dbp, pfef, true)) {
      warn << "cannot create " << dbp << "\n";
      return err;
    }
  }

  for (unsigned i=0; i < remove_path.size (); i++) {
    assert (!arg_checkonly);
    remove (remove_path [i]);
  }

  for (unsigned i = 0; i < add_path.size (); i++) {
    assert (!arg_checkonly);
    remove (add_path [i]);

    pfe e = add_pfe [i];
    if (e.cr_addr == 0)
      e.cr_addr = cr_addr;
    if (e.mtime.sec == 0 && e.mtime.nsec == 0)
      e.mtime = get_mtime (add_path [i]);
    if (insert (add_path [i], e))
      return -1;
  }

  for (unsigned i=0; i<adbs.size (); i++) {
    ptr<adb> d = adbs [i];
    d->closedb ();
  }

  for (unsigned i=0; i<opened_dirs.size (); i++) {
    struct stat sb;
    int r = lstat (opened_dirs [i], &sb);
    if (r == 0) {
      str f = get_info_file (opened_dirs [i]);
      if (btag == "") {
        btag = get_btag_from_infofile (f);
        if (!btag)
          btag = "";
      }
      str p;
      get_cfspath (opened_dirs [i], p);
      str m = p << "\n"
                << arg_logkeyfile << "\n"
	        << arg_projstr << "\n"
	        << arg_branch  << "\n"
	        << btag << "\n";
      if (!str2file (f, m, 0644, false)) {
        warn << "cannot write to " << f << "\n";
        return -1;
      }
    }
  }

  for (unsigned i=0; i<supers_dirs.size (); i++) {
    str f = get_supersedes_file (supers_dirs [i]);
    str buf = file2str (f);
    if (!buf) buf = "";
    buf = buf << armor64A (supers_commits [i]) << "\n";
    if (!str2file (f, buf, 0644, false)) {
      warn << "cannot write to " << f << "\n";
      return -1;
    }
  }

  return 0;
}

static void
split_path (str path, str &dir, str &name)
{
  dir = get_parent_dir (path, name);
}

int
pdb_mgr::insert (str path, pfe e)
{
  struct stat sb;
  if (lstat (path, &sb)) {
    warn << "cannot stat(3) " << path << "\n";
    return -1;
  }
  if ((sb.st_mode & S_IFDIR) == S_IFDIR) {
    ptr<adb> db = opendb (path);
    if (!db)
      return -1;
    ptr<pfe> ne = New refcounted <pfe> ();
    *ne = e;
    ne->name = "";
    db->insert ("", ne);
  }

  str dir;
  str name;
  split_path (path, dir, name);

  str p = get_info_file (dir);
  if (!lstat (p, &sb)) {
    ptr<adb> db = opendb (dir);
    if (!db)
      return -1;
    ptr<pfe> ne = New refcounted <pfe> ();
    *ne = e;
    ne->name = name;
    db->insert (name, ne);
  }
  return 0;
}

int
pdb_mgr::remove (str path)
{
  str dir;
  str name;
  split_path (path, dir, name);
 
  ptr<adb> db = opendb (dir);
  if (!db)
    return -1;
  db->del (name);
  return 0;
}

bool
pdb_mgr::exists (str path)
{
  str dir;
  str name;
  split_path (path, dir, name);
  
  ptr<adb> db = opendb (dir);
  if (!db)
    return false;
  ptr<adb_entry> v = db->lookup (name);
  if (v)
    return true;
  return false;
}

ptr<pfe>
pdb_mgr::lookup (str path)
{
  struct stat sb;
  
  if (!lstat (path, &sb) && (sb.st_mode & S_IFDIR) == S_IFDIR) {
    ptr<adb> db = opendb (path);
    if (db) {
      ptr<adb_entry> v = db->lookup ("");
      return *(reinterpret_cast <ptr<pfe> *> (&v));
    }
  }

  return lookup_in_parent_dir (path);
}

ptr<pfe>
pdb_mgr::lookup_in_parent_dir (str path)
{
  str dir;
  str name;
  split_path (path, dir, name);
  
  ptr<adb> db = opendb (dir);
  if (!db)
    return 0L;
  
  ptr<adb_entry> v = db->lookup (name);
  return *(reinterpret_cast <ptr<pfe> *> (&v));
}

int
pdb_mgr::remove_dir (str path)
{
  str f0 = get_ops_file (path);
  unlink (f0);
  str f1 = get_info_file (path);
  unlink (f1);
  str f2 = get_db_file (path);
  unlink (f2);
  str d0 = path << "/" << PASTWATCH_REF;
  rmdir (d0);
  if (rmdir (path))
    return -1;
  return 0;
}

