/*
 *
 * Copyright (C) 2003 Alexander Yip (yipal@lcs.mit.edu)
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include "admcontrol.h"
#include "chord_util.h"
#include "rxx.h"
#include "sfscrypt.h"
#include <stdio.h>
#include "misc_utils.h"

static rxx pkrx ("^0x([0-9a-fA-Fx]+)$");
static rxx hostlist_format ("^([\\s\\S]+?)\n(Version\\s+(\\d+)\\s*\n([\\s\\S]*))$"); 
static rxx hostlist_line ("^([0-9]{1,3}?\\.[0-9]{1,3}?\\.[0-9]{1,3}?\\.[0-9]{1,3}?)\\s+(\\d+)\\s+(0x[0-9a-fA-F]+)\\s+(\\S+)\\s*.*$");

hostlist::hostlist() : 
  raw_hostlist(""), hostlist_file(""),
  version(0), verifier(NULL) {
}

hostlist &
hostlist::only ()
{
  static ptr<hostlist> instance = NULL;
  if (instance == NULL)
    instance = New refcounted<hostlist> ();
  return *instance;
}

int 
hostlist::read_verifierpk(str filename) {
  str keyfile;
  if (keyfile = file2str (filename)) {
    if (keyfile.len() && keyfile[keyfile.len()-1] == '\n')
      keyfile = substr(keyfile, 0, keyfile.len()-1);
    if (verifier = sfscrypt.alloc(keyfile, SFS_VERIFY)) 
      return 1;
  }
  return 0;
}

int 
hostlist::verify_sig(str signedlist) {
  if( hostlist_format.match(signedlist)) {
    sfs_sig2 sig;
    if (str2sfs_sig2(hostlist_format[1], &sig)) {
      strbuf s;
      return verifier->verify(sig, hostlist_format[2]);
    } else
      return 0;
  }
  warn << "hostlist could not be parsed during verification\n";
  return 0;
}

int 
hostlist::parse_hostlist(str signedlist, int *out_version,
			 vec<chord_hostname> *names, vec<int32_t> *ports, 
			 vec<chordID> *ids, vec<ptr<sfspub> > *verifiers) {
  vec<str> lines;
  str list_part;
  int r, i;
  ptr<sfspub> pk;
  int status = 1;
  
  if( hostlist_format.match(signedlist)) {
    if (out_version)
      *out_version = atoi(hostlist_format[3]);

    list_part = hostlist_format[4];
    r = split(&lines, "\n", list_part);
    
    for(i=0; i<r; i++) {
      if (hostlist_line.match(lines[i]) && (pk = sfscrypt.alloc(hostlist_line[4], SFS_VERIFY))) {
	if (names)
	  names->push_back(hostlist_line[1]);
	if (ports)
	  ports->push_back(atoi(hostlist_line[2]));
	if (ids)
	  ids->push_back(bigint(hostlist_line[3]));
	if (verifiers)
	  verifiers->push_back(pk);
	
      } else {
	warn << "error parsing hostlist line " << 1 + i << ". Skipping line.\n";
	status = -1;
      }
    } 
    return status;
  }
  return 0;
}

int 
hostlist::read_hostlist(str filename) {
  hostentry *p;
  str signedhostlist;

  if (signedhostlist = file2str(filename)) {
    int v = verify_sig(signedhostlist);

    if (v && parse_hostlist(signedhostlist, NULL, NULL, NULL, NULL, NULL) != 0) {
      vec<chord_hostname> names;
      vec<int32_t> ports;
      vec<chordID> ids;
      vec<ptr<sfspub> > pubs;

      if (parse_hostlist(signedhostlist, &version, &names, &ports, &ids, &pubs)) {
	for(unsigned i=0; i<names.size(); i++){
	  p = insert_host(names[i], ports[i], ids[i], pubs[i]);
	}
	raw_hostlist = signedhostlist;
	hostlist_file = filename;
	return 1;
      }

    } else {
      warn << "Could not verify hostlist file: " << filename  << "\n";
      return 0;
    }
  }
  warn << "Could not read hostlist file: " << filename << "\n";
  return 0;
}

int
hostlist::verify_hostlist(str signedhostlist) {
  int version;

  if (!verify_sig(signedhostlist)) 
    return -1;
  if (parse_hostlist(signedhostlist, &version, NULL, NULL, NULL, NULL) < 1)
    return -1;

  return version;
}

void 
hostlist::write_hostlist(str newhostlist) {
  int v;
  if ((v = verify_hostlist(newhostlist))) {
    warn << "Updating hostlist to version " << v << "\n";
    str2file(hostlist_file, newhostlist);
  }
}

hostentry *hostlist::first() {
  return hosts.first();
}
hostentry *hostlist::next_circular(hostentry *current) {
  if (!current || current == hosts.last())
    return first();
  return hosts.next(current);
}

hostentry *hostlist::next(hostentry *current) {
  if (!current)
    return first();
  return hosts.next(current);
}

struct hostentry *
hostlist::find (chordID id) {
  return hosts.search(id);
}

struct hostentry *
hostlist::find (str name, int32_t port) {
  hostentry *p;
  p = hosts.first();
  while(p) {
    if ((p->name == name) &&
	(p->port == port))
      return p;
    
    p = hosts.next(p);
  }
  return NULL;
}

hostentry * 
hostlist::insert_host (chord_hostname name, int32_t port, chordID id, ptr<sfspub> pub) {
  hostentry *p;
  if ( (p = find(id)) ) {
    warn << "hostlist::insert_host: host id: " << id << " already exists.\n";
    return p;
  }

  if ( (p = find(name, port))) {
    warn << "hostlist::insert_host: host ip:port: " << name << ":" << port << " already exists.\n";
    return p;
  }

  struct hostentry *h = New hostentry();
  h->name = name;
  h->port = port;
  h->id = id;
  h->verifier = pub;

  hosts.insert(h);
  return h;
}

bool
hostlist::exists (chord_hostname name, int32_t port, chordID id, int vnode) {
  if (!initialized())
    return is_authenticID(id, name, port, vnode);

  return find(id);
}

chordID
hostlist::get_chordID (str addr, int port, int vnode)
{
  hostentry *p;
  if (!initialized()) 
    return make_chordID(addr, port, vnode);

  p = hosts.first();
  while(p) {
    if ((p->name == addr) &&
	(p->port == port))
      return p->id;
    
    p = hosts.next(p);
  }

  warn << "get_chordID could not find ID: " << addr << ":" << port << "\n";
  p = hosts.first();
  while(p) {
    warn << "   listed hosts: " << p->name << ":" << p->port << "\n";
    p = hosts.next(p);
  }

  return 0;
}

bool
hostlist::initialized () 
{
  return ( (bool)verifier && (bool)hosts.size() );
}

void
hostlist::set_myid( chordID id ) 
{
  my_id = id;
}

chordID
hostlist::get_myid()
{
  return my_id;
}

